/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
**  KOJAK       http://www.fz-juelich.de/jsc/kojak/                        **
*****************************************************************************
**  Copyright (c) 1998-2010                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  Copyright (c) 2003-2008                                                **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/
/**
* \file cube_score.cpp
* \brief Collect the data about every region in the cube and prints it out. 
Does same like "cube_regioninfo.cpp", a little extended verions. 
*/


#ifndef CUBE_COMPRESSED
#  include <fstream>
#else
#  include "zfstream.h"
#endif
#include <iomanip>
#include <string>
#include <set>
#include <map>
#include <numeric>
#include <vector>
#include <unistd.h>

#include "Blacklist.h"
#include "AggrCubeMaps.h"
#include "Cnode.h"
#include "Thread.h"
#include "Metric.h"
#include "Region.h"

#include "CnodeInfo.h"
#include "RegionInfo.h"

#include "helper.h"
#include "services.h"

using namespace std;
using namespace cube;
using namespace services;

string lowercase(const string& str);
/**
* A collection of all costs (??) for a region.
*/
struct tracefile_costs
{
    vector<unsigned long long> acc_costs_by_region;
    vector<unsigned long long> max_costs_by_region;
  
    unsigned long long acc_all;
    unsigned long long acc_bl;
    unsigned long long acc_com;
    unsigned long long acc_mpi;
    unsigned long long acc_omp;
    unsigned long long acc_usr;

    unsigned long long acc_com_bl;
    unsigned long long acc_mpi_bl;
    unsigned long long acc_omp_bl;
    unsigned long long acc_usr_bl;

    vector<unsigned long long> pt_all;
    vector<unsigned long long> pt_bl;
    vector<unsigned long long> pt_wbl;
    vector<unsigned long long> pt_com;
    vector<unsigned long long> pt_mpi;
    vector<unsigned long long> pt_omp;
    vector<unsigned long long> pt_usr;
    vector<unsigned long long> pt_com_bl;
    vector<unsigned long long> pt_mpi_bl;
    vector<unsigned long long> pt_omp_bl;
    vector<unsigned long long> pt_usr_bl;

    unsigned long long max_all;
    unsigned long long max_bl;
    unsigned long long max_wbl;
    unsigned long long max_com;
    unsigned long long max_mpi;
    unsigned long long max_omp;
    unsigned long long max_usr;
    unsigned long long max_com_bl;
    unsigned long long max_mpi_bl;
    unsigned long long max_omp_bl;
    unsigned long long max_usr_bl;

    tracefile_costs(size_t rcnt, size_t tcnt)
	: acc_costs_by_region(rcnt, 0),
	  max_costs_by_region(rcnt, 0),
	  pt_all(tcnt, 0),
	  pt_bl(tcnt, 0),
	  pt_wbl(tcnt, 0),
	  pt_com(tcnt, 0),
	  pt_mpi(tcnt, 0),
	  pt_omp(tcnt, 0),
	  pt_usr(tcnt, 0),
	  pt_com_bl(tcnt, 0),
	  pt_mpi_bl(tcnt, 0),
	  pt_omp_bl(tcnt, 0),
	  pt_usr_bl(tcnt, 0)

	{
	    acc_all=0;
	    acc_bl=0;
	    acc_com=0;
	    acc_mpi=0;
	    acc_omp=0;
	    acc_usr=0;
	    acc_com_bl=0;
	    acc_mpi_bl=0;
	    acc_omp_bl=0;
	    acc_usr_bl=0;
	    max_all=0;
	    max_bl=0;
	    max_wbl=0;
	    max_com=0;
	    max_mpi=0;
	    max_omp=0;
	    max_usr=0;
	    max_com_bl=0;
	    max_mpi_bl=0;
	    max_omp_bl=0;
	    max_usr_bl=0;
	}
/**
* A method of a collection, allows to get the maximum values for all kinds of regions.
*/
    void calculate_maxima()
	{
	    max_all=find_max(pt_all).first;
	    max_bl=find_max(pt_bl).first;
	    max_wbl=find_max(pt_wbl).first;
	    max_com=find_max(pt_com).first;
	    max_mpi=find_max(pt_mpi).first;
	    max_omp=find_max(pt_omp).first;
	    max_usr=find_max(pt_usr).first;
	    max_com_bl=find_max(pt_com_bl).first;
	    max_mpi_bl=find_max(pt_mpi_bl).first;
	    max_omp_bl=find_max(pt_omp_bl).first;
	    max_usr_bl=find_max(pt_usr_bl).first;
	}
};

/**
* A STL conform way to sort tbc values.
*/ 
class SortByMaxtbc : public std::binary_function<uint32_t, uint32_t, bool>
{
  public:
    SortByMaxtbc(const tracefile_costs& costs)
      : m_costs(costs)
    {}

    bool operator()(const uint32_t& lhs, const uint32_t& rhs) const
    {
      return m_costs.max_costs_by_region[lhs] > m_costs.max_costs_by_region[rhs];
    }

  private:
    const tracefile_costs& m_costs;
};

/// @cond PROTOTYPES
tracefile_costs calculate_tracefile_costs(AggrCubeMaps*, const  CRegionInfo&, CBlacklist*);
/// @endcond

/**
* A collection about all data for a given metric.
*/
template<typename T>
struct metric_data
{
    Metric* metric;
    std::string name;
    AggrCubeMaps* acube;
    T bl;
    T com;
    T com_bl;
    T mpi;
    T mpi_bl;
    T omp;
    T omp_bl;
    T total;
    T usr;
    T usr_bl;
    map<Region*, double> excl;
    map<Region*, double> incl;
    map<Region*, double> diff;

    metric_data()
	{
	    bl=T(0);
	    com=T(0);
	    com_bl=T(0);
	    mpi=T(0);
	    mpi_bl=T(0);
	    omp=T(0);
	    omp_bl=T(0);
	    total=T(0);
	    usr=T(0);
	    usr_bl=T(0);
	    metric=0;
	    acube=0;
	};
/**
* Returns a type of the metric value.
*/

    bool isfloat()
	{
	    if (metric!=0)
		return (metric->get_dtype() == "FLOAT");
	    else
		return false;
	}
};
/**
*  Calculates a percent.
*/
template<class S, class T>
double percent(S total, T costs)
{
    if (total!=0)
    {
	return 100.0 * (double) costs / (double) total;
    }
    else return 0.0;
}
/**
* Prints a usage of this tool.
*/
void usage(const char* filename)
{
    cout << "Usage: " << filename << " [-r] [-f <filename>] [-m <metric>[,<metric>...]] <cubefile>\n"
         << "  -r     Print metrics for each region\n"
#ifdef FILTER_FORMAT
         << "  -u     Mark all USR computation regions as filtered\n"
#endif
         << "  -f     File containing names of regions to mark as filtered\n"
         << "  -m     List of metrics that should be displayed, default: max_tbc,time\n\n"
         << "Regions are classified into the categories ANY (aggregate of all regions),\n"
	 << "MPI (pure MPI library functions), OMP (pure OpenMP functions/regions),\n"
         << "COM (\"combined\" user regions calling OpenMP/MPI, directly or indirectly), and\n"
         << "USR (user regions with purely local computation not containing MPI or OpenMP).\n\n"
         << "Metrics can be any of those defined in the cubefile, or two special metrics:\n"
         << "The total_tbc metric provides an estimate of total size of trace data (in bytes).\n"
	 << "The max_tbc metric provides an estimate for the trace buffer capacity (in bytes)\n"
	 << "that is required to store all events that would be generated by a single process.\n";
    exit(0);
}
/**
* Accumulate usr, mpi, com and total values in md.
*/
template<class T>
void acc_with_type(metric_data<T>& md, CRegionInfo& reginfo, CBlacklist* blacklist)
{
    /*
      accumulate usr, mpi, omp, com and total values in md
    */
    typename std::map<Region*, double>::const_iterator it;
    for (it = md.excl.begin(); it!=md.excl.end(); it++)
    {
	T nv((T)it->second);
	const Region* region(it->first);
      
	md.total += nv;
	switch (reginfo[region->get_id()])
	{
	case MPI:
	    md.mpi+=nv;
	    break;

	case OMP:
	    md.omp+=nv;
	    break;

	case COM:
	    md.com+=nv;
	    break;
	  
	case USR:
	    md.usr+=nv;
	    break;
	default:  // another types of regions are ignored
	    break;
	}

	if (blacklist!=0)
        {
	    if ((*blacklist)(region->get_id()))
	    {
		md.bl+=nv;
	    }
	    else
		switch (reginfo[region->get_id()])
		{
		case MPI:
		    md.mpi_bl+=nv;
		    break;

		case OMP:
		    md.omp_bl+=nv;
		    break;

		case COM:
		    md.com_bl+=nv;
		    break;
	
		case USR:
		    md.usr_bl+=nv;
		    break;
		default:// another types of regions are ignored
		    break;
		}
        }
    }
}
/**
* Main program. 
* - Check calling arguments
* - Read the .cube input file.
* - Create blacklist for the cube and regions in the cube.
* - Run over metrics (comma separated values in command parameter.) and for every metric do.
* - Calculate a values for every region for given metric.
* - Create a trace cost.
* - Prints the result in human readable form.
* - end.
*/

int main(int argc, char* argv[])
{
    string cubefile;
    string blacklistname;
    string metriclist;
    
    bool do_regions(false);
    bool do_blacklist(false);
    bool do_metriclist(false);
    bool do_maxtbc(false);
    bool do_totaltbc(false);
    
    int c;

    if (argc == 1)
	usage(argv[0]);
  
    opterr = 0;
    int mna(1);
  
    while ((c = getopt (argc, argv, "m:hruf:")) != -1)
    {
	switch (c)
	{
	case 'h':
	    usage(argv[0]);
	    break;
	case 'r':
	    do_regions=true;
	    mna++;
	    break;
        case 'u':
	    do_regions=true;
	    do_blacklist=true;
            if (!blacklistname.empty())
                cerr << "Warning: filter \"" << blacklistname 
                     << "\" will be ignored!" << endl;
            blacklistname="<USR>";
            mna++;
            break;
	case 'f':
	    blacklistname=string(optarg);
	    do_blacklist=true;
	    mna+=2;
	    break;
	case 'm':
	    metriclist=string(optarg);
	    do_metriclist=true;
	    mna+=2;
	    break;
	}
    }


	
    if (mna == (argc))
	usage(argv[0]);
    else if (mna == (argc-1))
	cubefile=string(argv[argc-1]);

    if (check_file(cubefile.c_str()) != 0 ) exit(-1);

    AggrCubeMaps *input = new AggrCubeMaps();

#ifndef CUBE_COMPRESSED
    ifstream in (cubefile.c_str());
#else
    gzifstream in (cubefile.c_str(), ios_base::in|ios_base::binary);
#endif
    // Read input file
    cerr << "Reading " << cubefile << "...";
    if (!in)
    {
	cerr << "Error: open " << cubefile << endl;
	exit(1);
    }
    in >> *input;
    cerr << " done." << endl;
  
    const vector<Region*>&  regions = input->get_regv();
  
    const size_t rcnt = regions.size();
    CRegionInfo reginfo = input;
  
    CBlacklist* blacklist = 0;
    if (do_blacklist)
    {
        if (blacklistname == "<USR>") {
            cerr << "Filtering all USR regions:" << endl;
            blacklist=new CBlacklist(input, &reginfo);
        } else {
            cerr << "Applying filter \"" << blacklistname << "\":" << endl;
            blacklist=new CBlacklist(input, blacklistname, &reginfo);
        }
    }

    /*
      BEGIN: pre-calculate total runtime and total number of visits
    */

    vector<metric_data<double> > metricdata;
    
    if (do_metriclist)
    {
	size_t oldpos=0;
	while (true)
	{
	    const size_t f = metriclist.find_first_of(',', oldpos);
	    string metricname;
	    if (f == string::npos)
	    {
		metricname=metriclist.substr(oldpos);
	    }
	    else
	    {
		metricname=metriclist.substr(oldpos, f-oldpos);

	    }

	    oldpos=f+1;
	    
	    if (metricname == "max_tbc")
	    {
		
		do_maxtbc=true;
		if (f == string::npos)
		    break;
		else
		    continue;
	    }

	    if (metricname == "total_tbc")
	    {
		do_totaltbc=true;
		if (f == string::npos)
		    break;
		else
		    continue;
	    }
	    metricdata.push_back(metric_data<double>());
	    metricdata[metricdata.size()-1].name=metricname;
	    if (f == string::npos) break;
	}
    }
    else
    {
	metricdata.push_back(metric_data<double>());
	metricdata[0].name=string("time");
	do_maxtbc=true;
    }

    for (size_t i=0; i<metricdata.size(); i++)
    {
	metricdata[i].metric=input->get_met(metricdata[i].name);
	
	if (metricdata[i].metric!=0)
	    input->get_reg_tree(metricdata[i].excl, metricdata[i].incl, metricdata[i].diff, INCL, INCL, metricdata[i].metric, 0);
	else
	{
            const vector<Metric*>& metrics = input->get_metv();
	    cout << "unknown metric: " << metricdata[i].name << endl;
            cout << "available metrics: " << endl;
            cout << "total_tbc (bytes) \"Total trace event byte count for all processes\"" << endl;
            cout << "max_tbc (bytes) \"Maximum trace event byte count for a process\"" << endl;
            for (size_t i=0; i<metrics.size(); i++) {
                Metric* met = metrics[i];
                cout << met->get_uniq_name() << " (" << met->get_uom() << ") \"";
                cout << met->get_disp_name();
                while (NULL != (met = met->get_parent()))
                    cout << " - " << met->get_disp_name();
                cout << "\"" << endl;
            }
	    exit(EXIT_FAILURE);
	}
  
	acc_with_type<double>(metricdata[i], reginfo, blacklist);
    }
    
    /*
      DONE: pre-calculate total runtime and total number of visits
    */
    
    tracefile_costs tbcosts(calculate_tracefile_costs(input, reginfo, blacklist));
  
    /*
      BEGIN: Print results
    */

    if (tbcosts.acc_omp > 0)
    {
      std::cerr << "Warning: Detected use of OpenMP. Calculated trace buffer\n"
                << "         requirements are likely to be wrong!\n\n";
    }

    cerr << "Estimated aggregate size of event trace (total_tbc): "
         << tbcosts.acc_all-tbcosts.acc_bl << " bytes" << endl;
    cerr << "Estimated size of largest process trace (max_tbc):   " 
         << tbcosts.max_all-tbcosts.max_bl << " bytes" << endl;

    cerr << "(Hint: When tracing set ELG_BUFFER_SIZE > max_tbc to avoid intermediate flushes";
    if (!do_blacklist) {
        cerr << endl << " or reduce requirements using file listing names of USR regions to be filtered";
    }
    cerr << ".)" << endl << endl;

    size_t w0 = 12;
  
    if (do_regions)
	/* calculate layout information */
	for (size_t i=0; i<rcnt; i++)
	{
	    size_t l = regions[i]->get_name().length();
	    if (l>w0) w0=l;
	}
    
    w0+=2;
    const int w1 = 12;
    const int w2 = 8;
  
#ifndef FILTER_FORMAT
    cout << right
	 << setw(3)  << "flt"
	 << setw(6)  << "type";
#else
    cout << "# " << left << setw(6) << "type" << setw(41) << "region" << right;
#endif
    if (do_maxtbc)
	cout << setw(w1) << "max_tbc";
    if (do_totaltbc)
	cout << setw(w1) << "total_tbc";

    for (size_t i=0; i<metricdata.size(); i++)
    {
	string::size_type n = string::npos;
	int sz = metricdata[i].name.length();
	if (sz > w1)
	    n=w2-3;
	
	cout << setw(w1) << metricdata[i].name.substr(0, n)
	     << setw(w2)  << "%";
    }

#ifndef FILTER_FORMAT
    cout << left
	 << " "
	 << "region";
#endif
    cout << endl;


#ifndef FILTER_FORMAT
    cout << right << (do_blacklist ? " - " : "   ") << "   ANY";
#else
    cout << left << setw(48) << "# ANY # (aggregation of all types)" << "#" << right;
#endif

    if (do_maxtbc)
	cout << setw(w1) << tbcosts.max_all;
    if (do_totaltbc)
	cout << setw(w1) << tbcosts.acc_all;
    for (size_t i=0; i<metricdata.size(); i++)
    {
	int prec = 0;
	if (metricdata[i].isfloat()) prec=2;	
	cout << setw(w1) << setprecision(prec) << fixed << metricdata[i].total
	     << setw(w2) << setprecision(2) << fixed << percent(metricdata[i].total, metricdata[i].total);
    }
#ifndef FILTER_FORMAT
    cout << left  << " (summary) ALL";
#endif
    cout << endl;

    if (tbcosts.acc_mpi > 0)
    {
#ifndef FILTER_FORMAT
      cout << right << (do_blacklist ? " - " : "   ") << "   MPI";
#else
      cout << left << setw(48) << "# MPI # (MPI library routines)" << "#" << right;
#endif
      if (do_maxtbc)
	  cout << setw(w1) << tbcosts.max_mpi;
      if (do_totaltbc)
	  cout << setw(w1) << tbcosts.acc_mpi;
      for (size_t i=0; i<metricdata.size(); i++)
      {
	  int prec = 0;
	  if (metricdata[i].isfloat()) prec=2;
	  cout << setw(w1) << setprecision(prec) << fixed << metricdata[i].mpi
	       << setw(w2) << setprecision(2) << fixed << percent(metricdata[i].total, metricdata[i].mpi);
      }
#ifndef FILTER_FORMAT
      cout << left  << " (summary) MPI";
#endif
      cout << endl;
    }

    if (tbcosts.acc_omp > 0)
    {
#ifndef FILTER_FORMAT
      cout << right << (do_blacklist ? " - " : "   ") << "   OMP";
#else
      cout << "#" << setw(44) << right << "(summary) # OMP";
      cout << left << setw(48) << "# OMP # (OpenMP API + parallel regions)" << "#" << right;
#endif
      if (do_maxtbc)
	  cout << setw(w1) << tbcosts.max_omp;
      if (do_totaltbc)
	  cout << setw(w1) << tbcosts.acc_omp;
      for (size_t i=0; i<metricdata.size(); i++)
      {
	  int prec = 0;
	  if (metricdata[i].isfloat()) prec=2;
	  cout << setw(w1) << setprecision(prec) << fixed << metricdata[i].omp
	       << setw(w2) << setprecision(2) << fixed << percent(metricdata[i].total, metricdata[i].omp);
      }
#ifndef FILTER_FORMAT
      cout << left  << " (summary) OMP";
#endif
      cout << endl;
    }

#ifndef FILTER_FORMAT
    cout << right << (do_blacklist ? " - " : "   ") << "   COM";
#else
    cout << left << setw(48) << "# COM # (combined OMP/MPI & USR computation)" << "#" << right;
#endif

    if (do_maxtbc)
	cout << setw(w1) << tbcosts.max_com;
    if (do_totaltbc)
	cout << setw(w1) << tbcosts.acc_com;
    
    for (size_t i=0; i<metricdata.size(); i++)
    {
	int prec = 0;
	if (metricdata[i].isfloat()) prec=2;
	
	cout << setw(w1) << setprecision(prec) << fixed << metricdata[i].com
	     << setw(w2) << setprecision(2) << fixed << percent(metricdata[i].total, metricdata[i].com);
    }
#ifndef FILTER_FORMAT
    cout << left << " (summary) COM";
#endif
    cout << endl;

#ifndef FILTER_FORMAT
    cout << right << (do_blacklist ? " - " : "   ") << "   USR";
#else
    cout << left << setw(48) << "# USR # (user computation routines)" << "#" << right;
#endif

    if (do_maxtbc)
	cout << setw(w1) << tbcosts.max_usr;
    if (do_totaltbc)
	cout << setw(w1) << tbcosts.acc_usr;
    
    for (size_t i=0; i<metricdata.size(); i++)
    {
	int prec = 0;
	if (metricdata[i].isfloat()) prec=2;
	
	cout << setw(w1) << setprecision(prec) << fixed << metricdata[i].usr
	     << setw(w2) << setprecision(2) << fixed << percent(metricdata[i].total, metricdata[i].usr);
    }
	
#ifndef FILTER_FORMAT
    cout << left  << " (summary) USR" << endl;
#endif
    cout << endl;

    if (do_blacklist)
    {
#ifndef FILTER_FORMAT
	cout << right << " + " << "   FLT";
#else
        cout << "# ";
        for (size_t i=2; i<=80; i++) cout << '-';
        cout << endl;
        cout << left << setw(48) << "# FLT # (eliminated with filter)" << "#" << right;
#endif
	if (do_maxtbc)
	    cout << setw(w1) << tbcosts.max_bl;
	if (do_totaltbc)
	    cout << setw(w1) << tbcosts.acc_bl;
	
	for (size_t i=0; i<metricdata.size(); i++)
	{
	    int prec = 0;
	    if (metricdata[i].isfloat()) prec=2;

	    cout << setw(w1) << setprecision(prec) << fixed << metricdata[i].bl
		 << setw(w2) << setprecision(2) << fixed << percent(metricdata[i].total, metricdata[i].bl);
	}
	    
#ifndef FILTER_FORMAT
	cout	<< left  << " (summary) FLT";
	cout	<< endl;
	cout << right << " * " << "   ANY";
#else
	cout	<< endl;
        cout << left << setw(48) << "# ANY # (remaining after filter)" << "#" << right;
#endif

	if (do_maxtbc)
	    cout << setw(w1) << tbcosts.max_wbl;
	if (do_totaltbc)
	    cout << setw(w1) << tbcosts.acc_all-tbcosts.acc_bl;
	
	for (size_t i=0; i<metricdata.size(); i++)
	{
	    int prec = 0;
	    if (metricdata[i].isfloat()) prec=2;

	    cout << setw(w1) << setprecision(prec) << fixed << metricdata[i].total-metricdata[i].bl
		 << setw(w2) << setprecision(2) << fixed << percent(metricdata[i].total, (metricdata[i].total-metricdata[i].bl));
	}
	    
#ifndef FILTER_FORMAT
	cout	<< left  << " (summary) ALL-FLT";
#endif
	cout	<< endl;

        if (tbcosts.acc_mpi > 0)
        {
#ifndef FILTER_FORMAT
	  cout << right << " - " << "   MPI";
#else
          cout << left << setw(48) << "# MPI # (remaining after filter)" << "#" << right;
#endif
	  if (do_maxtbc)
	      cout << setw(w1) << tbcosts.max_mpi_bl;
	  if (do_totaltbc)
	      cout << setw(w1) << tbcosts.acc_mpi_bl;

	  for (size_t i=0; i<metricdata.size(); i++)
	  {
	      int prec = 0;
	      if (metricdata[i].isfloat()) prec=2;

	      cout << setw(w1) << setprecision(prec) << fixed << metricdata[i].mpi_bl
		   << setw(w2) << setprecision(2) << fixed << percent(metricdata[i].total, metricdata[i].mpi_bl);
	  }
#ifndef FILTER_FORMAT
	  cout	<< left  << " (summary) MPI-FLT";
#endif
          cout  << endl;
        }

        if (tbcosts.acc_omp > 0)
        {
#ifndef FILTER_FORMAT
	  cout << right << " - " << "   OMP";
#else
          cout << left << setw(48) << "# OMP # (remaining after filter)" << "#" << right;
#endif
	  if (do_maxtbc)
	      cout << setw(w1) << tbcosts.max_omp_bl;
	  if (do_totaltbc)
	      cout << setw(w1) << tbcosts.acc_omp_bl;

	  for (size_t i=0; i<metricdata.size(); i++)
	  {
	      int prec = 0;
	      if (metricdata[i].isfloat()) prec=2;

	      cout << setw(w1) << setprecision(prec) << fixed << metricdata[i].omp_bl
		   << setw(w2) << setprecision(2) << fixed << percent(metricdata[i].total, metricdata[i].omp_bl);
	  }
#ifndef FILTER_FORMAT
	  cout	<< left  << " (summary) OMP-FLT";
#endif
          cout  << endl;
        }

#ifndef FILTER_FORMAT
	cout << right << " * " << "   COM";
#else
        cout << left << setw(48) << "# COM # (remaining after filter)" << "#" << right;
#endif

	if (do_maxtbc) cout << setw(w1) << tbcosts.max_com_bl;
	if (do_totaltbc) cout << setw(w1) << tbcosts.acc_com_bl;
	
	for (size_t i=0; i<metricdata.size(); i++)
	{
	    int prec = 0;
	    if (metricdata[i].isfloat()) prec=2;

	    cout << setw(w1) << setprecision(prec) << fixed << metricdata[i].com_bl
		 << setw(w2) << setprecision(2) << fixed << percent(metricdata[i].total, metricdata[i].com_bl);
	}
#ifndef FILTER_FORMAT
	cout	<< left  << " (summary) COM-FLT";
#endif
        cout    << endl;

#ifndef FILTER_FORMAT
	cout << right << " * " << "   USR";
#else
        cout << left << setw(48) << "# USR # (remaining after filter)" << "#" << right;
#endif
	if (do_maxtbc)
	    cout << setw(w1) << tbcosts.max_usr_bl;
	if (do_totaltbc)
	    cout << setw(w1) << tbcosts.acc_usr_bl;
	
	for (size_t i=0; i<metricdata.size(); i++)
	{
	    int prec = 0;
	    if (metricdata[i].isfloat()) prec=2;
	
	    cout << setw(w1) << setprecision(prec) << fixed << metricdata[i].usr_bl
		 << setw(w2) << setprecision(2) << fixed << percent(metricdata[i].total, metricdata[i].usr_bl);
	}
	
#ifndef FILTER_FORMAT
	cout	<< left  << " (summary) USR-FLT" << endl;
#endif
        cout << endl;
    }

    if (do_regions)
    {
#ifdef FILTER_FORMAT
      cout << "# ";
      for (size_t i=2; i<=80; i++) cout << '-';
      cout << endl;
#endif

      // Sort regions by trace buffer costs
      vector<uint32_t> region_ids;
      region_ids.reserve(rcnt);
      for (size_t i=0; i<rcnt; i++)
        region_ids.push_back(i);
      sort(region_ids.begin(), region_ids.end(), SortByMaxtbc(tbcosts));

      for (size_t k=0; k<rcnt; k++)
      {
        int i = region_ids[k];

        if (reginfo[i]==NUL) continue;
        if (tbcosts.acc_costs_by_region[i] == 0) continue;

#ifndef FILTER_FORMAT
        if (do_blacklist)
        {
	        if ((*blacklist)((Region *)regions[i]))
	        cout << " + ";
	        else
	        cout << " - ";
        }
        else
	        cout << "   ";

        cout << right
	         << setw(6) << Callpathtype2String(reginfo[i]);
        if (do_maxtbc)
	        cout << setw(w1) << tbcosts.max_costs_by_region[i];
        if (do_totaltbc)
	        cout << setw(w1) << tbcosts.acc_costs_by_region[i];

        for (size_t j=0; j<metricdata.size(); j++)
        {
	        int prec = 0;
	        if (metricdata[j].isfloat()) prec=2;
	        cout << setprecision(prec) << fixed
	         << setw(w1) << metricdata[j].excl[regions[i]]
	         << setprecision(2) << fixed
	         << setw(w2) << percent(metricdata[j].total, metricdata[j].excl[regions[i]]);
        }

        cout << left
	         << " "
	         << regions[i]->get_name() << endl;
#else // filter-format
        if (!do_blacklist || !(*blacklist)((Region *)regions[i])) {
            cout << "# " << Callpathtype2String(reginfo[i]);
            // distinguish non-filterable regions
            if ((reginfo[i] == EPK) || (reginfo[i] == MPI) || (reginfo[i] == OMP)) cout << " #";
        }
        cout << '\t';

        cout << setw(39) << left << regions[i]->get_name();

        cout << "\t#" << right;
        if (do_maxtbc)
	        cout << setw(w1) << tbcosts.max_costs_by_region[i];
        if (do_totaltbc)
	        cout << setw(w1) << tbcosts.acc_costs_by_region[i];

        for (size_t j=0; j<metricdata.size(); j++)
        {
	        int prec = metricdata[j].isfloat() ? 2 : 0;
	        cout << setprecision(prec) << fixed
	         << setw(w1) << metricdata[j].excl[regions[i]]
	         << setprecision(2) << fixed
	         << setw(w2) << percent(metricdata[j].total, metricdata[j].excl[regions[i]]);
        }

        cout << endl;
#endif
      }
    }

/* END: Print results */

    delete input;
}
/**
* Calculate a cost ( given as a severity value mulpilied wih a factor ) for every 
region for every thread. Calculate a cost for every kind of regions 
(usr, mpi, omp, and so on) over all threads. And get a maximum value at the end.
 */
tracefile_costs calculate_tracefile_costs(AggrCubeMaps* input, const CRegionInfo& reginfo, CBlacklist* blacklist)
{   
    const vector<Region*>&  regions = input->get_regv();
    const vector<Thread*>&  threads = input->get_thrdv();
    
    const size_t rcnt = regions.size();
    const size_t tcnt = threads.size();

    tracefile_costs tbcosts = tracefile_costs(rcnt, tcnt);

    Metric* visits = input->get_met("visits");	

    /* accumulated costs on a per thread basis */
    for (size_t i=0; i<rcnt; i++)
    {
	const Region* region = regions[i];
	const unsigned long long d = TypeFactor(region->get_name());
	const vector<Cnode*>& cnodev = region->get_cnodev();
	const size_t ccnt = cnodev.size();
	vector<unsigned long long> buffer(tcnt,0);

	for (size_t j=0; j<ccnt; j++)
	{
	    Cnode* cnode = cnodev[j];
	    for (size_t ThreadId=0; ThreadId<tcnt; ThreadId++)
	    {
		Thread* thread = threads[ThreadId];

		const unsigned long long newval=static_cast<unsigned long long>(input->get_sev(visits, cnode, thread));
		const unsigned long long costs=newval*d;
		buffer[ThreadId]+=costs;
		tbcosts.pt_all[ThreadId]+=costs;
		tbcosts.acc_costs_by_region[i]+=costs;
		if (blacklist!=0)
		{
		    if ((*blacklist)((Region*) region))
		    {
			tbcosts.pt_bl[ThreadId]+=costs;
		    }
		    else
		    {
			switch (reginfo[i])
			{
			case MPI:
			    tbcosts.pt_mpi_bl[ThreadId]+=costs;
			    break;

			case OMP:
			    tbcosts.pt_omp_bl[ThreadId]+=costs;
			    break;

			case USR:
			    tbcosts.pt_usr_bl[ThreadId]+=costs;
			    break;
	      
			case COM:
			    tbcosts.pt_com_bl[ThreadId]+=costs;
			    break;
			default:// another types of regions are ignored
			    break;
			}
			tbcosts.pt_wbl[ThreadId]+=costs;
		    }
		}

		switch (reginfo[i])
		{
		case MPI:
		    tbcosts.pt_mpi[ThreadId]+=costs;
		    break;

		case OMP:
		    tbcosts.pt_omp[ThreadId]+=costs;
		    break;

		case USR:
		    tbcosts.pt_usr[ThreadId]+=costs;
		    break;
		  
		case COM:
		    tbcosts.pt_com[ThreadId]+=costs;
		    break;
		default:// another types of regions are ignored
		    break;
		}
	    }
	  
	}

	unsigned long long max_costs=find_max(buffer).first;
	if ( tbcosts.max_costs_by_region[i] < max_costs)
	    tbcosts.max_costs_by_region[i]=max_costs;	
    }

    /*
      BEGIN: calculate total costs split by category (mpi, omp, com, usr, blacklist) using get_met_tree(...)
    */

    for (size_t regionId=0; regionId<regions.size(); regionId++)
    {
	Region* region=regions[regionId];
	const vector<Cnode*>& cnodev(region->get_cnodev());
	unsigned long long d=TypeFactor(region->get_name());
	for (size_t cnodeId=0; cnodeId<cnodev.size(); cnodeId++)
	{
	    map<Metric*, double> excl_metrics;
	    map<Metric*, double> incl_metrics;
	    Cnode* cnode = cnodev[cnodeId];
	    input->get_met_tree(excl_metrics, incl_metrics, EXCL, INCL, cnode, 0);
	    Metric* metric = input->get_met("visits");
	    
	    unsigned long long visits((unsigned long long )excl_metrics[metric]);
	    unsigned long long nc = d*visits;
	    
	    tbcosts.acc_all+=nc;
      
	    if (blacklist!=0)
            {
		if ((*blacklist)((Region*) region))
		{
		    tbcosts.acc_bl+=nc;
		}
		else
                {
		    switch (reginfo[regionId])
		    {
		    case MPI:
			tbcosts.acc_mpi_bl+=nc;
			break;
		    case OMP:
			tbcosts.acc_omp_bl+=nc;
			break;
		    case USR:
			tbcosts.acc_usr_bl+=nc;
			break;
		    case COM:
			tbcosts.acc_com_bl+=nc;
			break;
		    default:// another types of regions are ignored
			break;
		    }
                }
            }

	    switch (reginfo[regionId])
	    {
	    case MPI:
		tbcosts.acc_mpi+=nc;
		break;
	    case OMP:
		tbcosts.acc_omp+=nc;
		break;
	    case USR:
		tbcosts.acc_usr+=nc;
		break;
	    case COM:
		tbcosts.acc_com+=nc;
		break;
	    default:// another types of regions are ignored
	        break;
	    }
	}
    }

    /*
      DONE: calculate total costs split by category (mpi, omp, com, usr, blacklist) using get_met_tree(...)
    */

    /*
      BEGIN: accumulute per thread data items
    */

    tbcosts.calculate_maxima();
  
    /*
      END: accumulute per thread data items
    */
  
    return tbcosts;
}
